/**************************************************************************
***    
*** Copyright (c) 1995-2000 Regents of the University of California,
***               Andrew E. Caldwell, Andrew B. Kahng and Igor L. Markov
*** Copyright (c) 2000-2002 Regents of the University of Michigan,
***               Saurabh N. Adya and Igor L. Markov
***
***  Contact author(s): abk@cs.ucsd.edu, imarkov@umich.edu
***  Original Affiliation:   UCLA, Computer Science Department,
***                          Los Angeles, CA 90095-1596 USA
***
***  Permission is hereby granted, free of charge, to any person obtaining 
***  a copy of this software and associated documentation files (the
***  "Software"), to deal in the Software without restriction, including
***  without limitation 
***  the rights to use, copy, modify, merge, publish, distribute, sublicense, 
***  and/or sell copies of the Software, and to permit persons to whom the 
***  Software is furnished to do so, subject to the following conditions:
***
***  The above copyright notice and this permission notice shall be included
***  in all copies or substantial portions of the Software.
***
*** THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, 
*** EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
*** OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
*** IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
*** CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT
*** OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR
*** THE USE OR OTHER DEALINGS IN THE SOFTWARE.
***
***
***************************************************************************/





#include "ABKCommon/abkcommon.h"
#include "RBPlacement.h"
#include "CongestionMaps/CongestionMaps.h"

int main(int argc, const char *argv[])
{
  NoParams         noParams       (argc,argv);
  BoolParam        helpRequest    ("help",argc,argv);
  BoolParam        helpRequest1   ("h",   argc,argv);
  BoolParam        remOverlaps    ("legal",argc,argv);
  BoolParam        snapToSites("snapToSites", argc, argv);

  RBPlacement::Parameters rbParams(argc,argv);
  StringParam   saveAsNodes("saveAsNodes", argc, argv);
  StringParam   saveAsCplace("saveAsCplace", argc, argv);
  StringParam   saveLEFDEF ("saveLEFDEF",  argc, argv);
  BoolParam     markMacrosAsBlocks("markMacrosAsBlocks", argc, argv);
  BoolParam   remSitesMacro("remSitesMacro", argc, argv);
  BoolParam   fixMacros("fixMacros", argc, argv);
  StringParam   saveAsNodesFloorplan("saveAsNodesFloorplan", argc, argv);
  StringParam   saveMacrosAsNodesFloorplan("saveMacrosAsNodesFloorplan", 
					   argc, argv);
  StringParam   saveAsNodesShredHW("saveAsNodesShredHW", argc, argv);
  StringParam   saveAsNodesWPinAssign("saveAsNodesWPinAssign", argc, argv);
  StringParam   savePlacementUnShred("savePlacementUnShred", argc, argv);
  StringParam   savePl("savePl", argc, argv);
  StringParam   plotNets("plotNets", argc, argv);
  StringParam   plotNodes("plotNodes", argc, argv);
  StringParam   plotNodesAndNets("plot", argc, argv);
  StringParam   plotNodesWNames("plotNodesWNames", argc, argv);
  StringParam   plotNodesAndNetsWNames("plotWNames", argc, argv);
  StringParam   plotSites("plotSites", argc, argv);
  StringParam   plotNodesWSites("plotNodesWSites", argc, argv);  
  StringParam   plotRows("plotRows", argc, argv);

  if (noParams.found() || helpRequest.found() || helpRequest1.found())
  {
     cout << "  Use '-help' or '-f filename.aux' " << endl;
     cout << "  Other options: \n"
	  << "   -xmin,-xmax,-ymin,-ymax  (for required range. Default means all)\n"
          << "   -remSitesMacro   (remove Sites below Macros)\n"
          << "   -fixMacros   (mark Macros as terminals while saving)\n"
          << "   -saveAsNodes     base_filename\n"
	  << "   -saveLEFDEF      base_filename\n"
          << "   -saveAsCplace    base_filename\n"
          << "   -markMacrosAsBlocks (for saveLEFDEF)\n"
	  << "   -savePl  filename\n"
          << "   -saveAsNodesWPinAssign     base_filename\n"
          << "   -saveAsNodesFloorplan base_filename\n"
          << "   -saveMacrosAsNodesFloorplan base_filename\n"
	  << "   -saveAsNodesShredHW  base_filename\n"
 	  << "   -savePlacementUnShred  filename\n"
	  << "   -plotNets        base_fileName\n"
          << "   -plotNodes       base_fileName\n"
          << "   -plotNodesWNames base_fileName\n"
          << "   -plot            base_fileName   (plots nodes and nets)\n"
          << "   -plotWNames      base_fileName   (the above + node names)\n"
	  << "   -plotSites	  base_fileName   (plots the site map)\n"
	  << "   -plotNodesWSites base_fileName   (the above + nodes)\n"
          << "   -plotRows        base_fileName   (plot the rows)\n"
          << "   -legal  (Use to remove any overlaps in existing placement)\n"
          << "   -snapToSites  (Use to snap cells to sites in existing placement)\n"
          << endl;
     return 0;
  }
  cout<<"RBPlacement Parameters"<<endl;
  cout<<rbParams<<endl;

  StringParam auxFile("f", argc, argv);
  abkfatal(auxFile.found(), "must have -f <auxfilename>");
  RBPlacement rbplace(auxFile, rbParams);


// add code that produces gnuplot files by traversing the hypergraph
// and the layout from rbplace   


  if(remOverlaps.found())
    {
      rbplace.snapCellsInSites();
      rbplace.remOverlaps();
    }

  if(snapToSites.found())
    rbplace.snapCellsInSites();

  if(remSitesMacro)
         rbplace.removeSitesBelowMacros();

  if(saveLEFDEF.found()) rbplace.saveLEFDEF(saveLEFDEF, markMacrosAsBlocks);

  if(saveAsNodes.found())
    {
      if(fixMacros.found())
	rbplace.saveAsNodesNetsWts(saveAsNodes, fixMacros);
      else
	rbplace.saveAsNodesNetsWts(saveAsNodes);
    }

  if(saveAsNodesWPinAssign.found())
    {
      rbplace.assignPinsToCells();
      rbplace.saveAsNodesNetsWts(saveAsNodesWPinAssign);
    }

  if(saveAsNodesFloorplan.found())
    rbplace.saveAsNodesNetsPlFloorplan(saveAsNodesFloorplan);

  if(saveMacrosAsNodesFloorplan.found())
    rbplace.saveMacrosAsNodesNetsPlFloorplan(saveMacrosAsNodesFloorplan);

  if(saveAsNodesShredHW.found()) 
    rbplace.saveAsNodesNetsWtsShredHW(saveAsNodesShredHW);

  if(savePlacementUnShred.found()) 
    rbplace.savePlacementUnShredHW(savePlacementUnShred);

  if(savePl.found()) 
    rbplace.savePlacement(savePl);

  if(saveAsCplace.found()) rbplace.saveAsCplace(saveAsCplace);

  double minx,maxx,miny,maxy;
  DoubleParam xmin("xmin",argc,argv);
  DoubleParam xmax("xmax",argc,argv);
  DoubleParam ymin("ymin",argc,argv);
  DoubleParam ymax("ymax",argc,argv);

  minx  = ( xmin.found() ?  xmin:-DBL_MAX);
  maxx  = ( xmax.found() ?  xmax:DBL_MAX);
  miny  = ( ymin.found() ?  ymin:-DBL_MAX);
  maxy  = ( ymax.found() ?  ymax:DBL_MAX);

   if(plotNets.found())
    	 rbplace.saveAsPlot("plotNets",minx,maxx,miny,maxy,plotNets);

   if(plotNodes.found())
    	 rbplace.saveAsPlot("plotNodes",minx,maxx,miny,maxy,plotNodes);

   if(plotNodesAndNets.found())
	 rbplace.saveAsPlot("plot",minx,maxx,miny,maxy,plotNodesAndNets);

  if(plotNodesWNames.found())
    	 rbplace.saveAsPlot("plotNodesWNames",minx,maxx,miny,maxy,plotNodesWNames);

   if(plotNodesAndNetsWNames.found())
	 rbplace.saveAsPlot("plotWNames",minx,maxx,miny,maxy,plotNodesAndNetsWNames);

   if(plotSites.found())
	 rbplace.saveAsPlot("plotSites",minx,maxx,miny,maxy,plotSites);

   if(plotNodesWSites.found())
         rbplace.saveAsPlot("plotNodesWSites",minx,maxx,miny,maxy,plotNodesWSites);

   if(plotRows.found())
	 rbplace.saveAsPlot("plotRows",minx,maxx,miny,maxy,plotRows);

   
   //cout<<endl<<endl<<auxFile<<" numMacros is : "<<rbplace.getNumMacros()<<endl<<endl;

  return 0;
}



